# Copyright 2023 The DLRover Authors. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import os
import pickle
import shutil
import signal
import threading
import time
from abc import ABCMeta, abstractmethod
from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Callable, Dict, List, Mapping, Optional, Tuple

import torch

from dlrover.python.common import env_utils
from dlrover.python.common.constants import CheckpointConstant, NodeEnv
from dlrover.python.common.log import default_logger as logger
from dlrover.python.common.multi_process import (
    SharedDict,
    SharedLock,
    SharedMemory,
    SharedQueue,
)

DLROVER_CKPT_CONFIG_KEY = "_DLORVER_CKPT_CONFIG"


class CheckpointSharedObjPrefix:
    SAVE_STEP_QNAME = "checkpoint_lock_rank_"
    META_NAME = "checkpoint_meta_"
    SHM_NAME = "checkpoint_shm_"
    SHM_LOCK_NAME = "shm_lock_"


class CheckpointEventType(Enum):
    SAVE = auto()
    UPDATE_SHARD = auto()
    EXIT = auto()


@dataclass
class CheckpointEvent:
    type: CheckpointEventType = CheckpointEventType.SAVE
    step: int = 0
    global_shard_num: int = 0


@dataclass
class SaverClassMeta:
    module_path: str = ""
    class_name: str = ""
    init_args: Dict[str, str] = None  # type: ignore


@dataclass
class TensorMeta:
    shape: Tuple[int] = None  # type: ignore
    dtype: torch.dtype = None  # type: ignore
    element_size: int = 0
    numel: int = 0
    offset: int = 0


@dataclass
class CheckpointShardConfig:
    """
    The configuration of a checkpointing shard on the training process.

    Attrbiutes:
        step (int): the global interation step.
        writing_shm (bool): the flag whether the training process is writing
            the state dict into the shared memory.
    """

    step: int = 0
    writing_shm: bool = False


@dataclass
class SingleFileCheckpointConfig(CheckpointShardConfig):
    """
    The configuration of a checkpointing shard to save the optimizer
    and model state dict into a file.

    Attrbiutes:
        path (str): the path to save the checkpoint shard.
    """

    path: str = ""


@dataclass
class DeepSpeedCheckpointConfig(CheckpointShardConfig):
    """
    The configuration of a checkpointing shard to save the DeepSpeed
    ZERO-0/1/2/3 stage.

    Attrbiutes:
        model_path (str): the path to save the checkpoint shard of model.
        optimizer_path (str): the path to save the checkpoint
            shard of optimizer.
    """

    model_path: str = ""
    optimizer_path: str = ""


def _traverse_state_dict(value: object, visitor: Callable[[object], None]):
    """
    Invoke ``visitor`` for each value recursively in ``state_dict``.
    """
    if isinstance(value, Mapping):
        temp_dict = {}
        for k, v in value.items():
            temp_dict[k] = _traverse_state_dict(v, visitor)
        return temp_dict
    elif isinstance(value, List):
        temp_list = []
        for _, v in enumerate(value):
            temp_list.append(_traverse_state_dict(v, visitor))
        return temp_list
    else:
        return visitor(value)


def _read_state_dict_from_shm(meta_dict, tensor_shm):
    state_dict = _traverse_state_dict(
        meta_dict,
        lambda x: _read_tensor_from_buf(x, tensor_shm),
    )
    return state_dict


def _read_tensor_from_buf(value, shm_tensor_buffer):
    """
    Read a tensor from the buffer of shared memory.
    """
    if isinstance(value, TensorMeta):
        if value.numel == 0:
            return torch.tensor([], dtype=value.dtype)
        else:
            shm_tensor = torch.frombuffer(
                buffer=shm_tensor_buffer.buf,
                dtype=value.dtype,
                offset=value.offset,
                count=value.numel,
            )
            value = shm_tensor.reshape(value.shape)
            return value
    else:
        return value


def _create_shared_memory(name, create, size=0):
    """
    Create a shared memory.
    """
    if not create:
        try:
            return SharedMemory(name=name)
        except FileNotFoundError:
            return None
    try:
        shm = SharedMemory(
            name=name,
            create=create,
            size=size,
        )
    except FileExistsError:
        shm = SharedMemory(name=name)
        if shm.size != size:
            logger.info("Re-create a new memory buffer.")
            shm.unlink()
            shm = SharedMemory(
                name=name,
                create=create,
                size=size,
            )
    return shm


def _traverse_copy_to_shm(value, meta, buffer):
    if isinstance(value, Mapping):
        for k, v in value.items():
            if isinstance(v, (Mapping, List)):
                m = meta[k]
                _traverse_copy_to_shm(v, m, buffer)
            elif torch.is_tensor(v):
                m = meta[k]
                _write_shared_memory(v, m, buffer)
            else:
                meta[k] = v
    elif isinstance(value, List):
        for i, v in enumerate(value):
            if isinstance(v, (Mapping, List)):
                m = meta[i]
                _traverse_copy_to_shm(v, m, buffer)
            elif torch.is_tensor(v):
                m = meta[i]
                _write_shared_memory(v, m, buffer)
            else:
                meta[i] = v


def _write_shared_memory(value: torch.Tensor, meta: TensorMeta, buffer):
    """
    Write a CPU tensor into the shared memory.
    """
    if value.numel() == 0:
        return
    shm_tensor = torch.frombuffer(
        buffer, dtype=value.dtype, count=value.numel(), offset=meta.offset
    ).reshape(value.shape)
    shm_tensor.copy_(value)


class SharedMemoryHandler(object):
    """
    The handler to write and read the shared memory with the state dict
    of PyTorch Module.

    Args:
        local_rank (int): the local rank of the process on a node.
        host (bool): the handler is on the host if True, otherwise,
            the handler is on the device.
    """

    def __init__(self, local_rank, host=True):
        self._buffer_size = 0
        meta_name = CheckpointSharedObjPrefix.META_NAME + str(local_rank)
        job_name = os.getenv(NodeEnv.TORCHELASTIC_RUN_ID, "")
        if job_name:
            self._shm_name = (
                job_name
                + "_"
                + CheckpointSharedObjPrefix.SHM_NAME
                + str(local_rank)
            )
        else:
            self._shm_name = CheckpointSharedObjPrefix.SHM_NAME + str(
                local_rank
            )
        self.shared_memory: Optional[SharedMemory] = None
        self.metadata = SharedDict(name=meta_name, create=host)

    def close(self):
        if self.shared_memory:
            self.shared_memory.close()

    def unlink(self):
        if not self.shared_memory:
            # The shared memory may be created by other processes.
            self.init_shared_memory()
        if self.shared_memory:
            self.shared_memory.unlink()
            logger.info(f"Unlink the shared memory {self._shm_name}")
        if self.metadata:
            self.metadata.unlink()

    def reset(self):
        self.close()
        if self.shared_memory:
            self.shared_memory.unlink()
        self.metadata.set({})
        self.shared_memory = None

    def _create_tensor_meta(self, value: torch.Tensor):
        """
        Create a tensor meta of a tensor and compute the total
        size of the state dict.
        """
        if not torch.is_tensor(value):
            return value
        meta = TensorMeta(
            shape=tuple(value.shape),  # type: ignore
            dtype=value.dtype,
            element_size=value.element_size(),
            numel=value.numel(),
            offset=self._buffer_size,
        )
        self._buffer_size += value.numel() * value.element_size()
        return meta

    def save_state_dict(self, state_dict, ckpt_conf: CheckpointShardConfig):
        """
        Copy the state dict from CPU memory buffer into the shared memory.
        """
        if not self.shared_memory:
            meta_dict = _traverse_state_dict(
                state_dict, self._create_tensor_meta
            )
            self.init_shared_memory(create=True, size=self._buffer_size)
        else:
            meta_dict = self.metadata.get(local=True)
        ckpt_conf.writing_shm = True
        meta_dict[DLROVER_CKPT_CONFIG_KEY] = ckpt_conf

        self.metadata.set(meta_dict)
        assert self.shared_memory is not None
        _traverse_copy_to_shm(state_dict, meta_dict, self.shared_memory.buf)
        ckpt_conf.writing_shm = False
        self.metadata.set(meta_dict)

    def load_state_dict(self):
        """
        Load the state dict from the shared memory.

        Returns:
            Tuple(int, dict): The first value is the iteration step,
                the second value is the state dict.
        """
        meta_dict = self.metadata.get()
        default_config = CheckpointShardConfig()
        config = meta_dict.get(DLROVER_CKPT_CONFIG_KEY, default_config)
        if not meta_dict or config.writing_shm:
            return {}
        if self.shared_memory is None:
            self.init_shared_memory(create=False)
        if not self.shared_memory:
            return {}

        state_dict = _read_state_dict_from_shm(meta_dict, self.shared_memory)
        state_dict.pop(DLROVER_CKPT_CONFIG_KEY, None)
        return state_dict

    def no_checkpint_state(self):
        """
        The handler lazily initializes the shared memory. The shared memory
        of the handler on the host may be None even if the handler on the
        device has saved state dict.
        """
        meta_dict = self.metadata.get()
        config: CheckpointShardConfig = meta_dict.get(
            DLROVER_CKPT_CONFIG_KEY, None
        )
        if config is None or config.step == 0:
            return True
        return False

    def init_shared_memory(self, create=False, size=0):
        self.shared_memory = _create_shared_memory(
            self._shm_name, create=create, size=size
        )

    def get_checkpoint_config(self, default_config):
        """
        Get the configuration of checkpointing state dict in the shared
        memory.

        Returns:
            A CheckpointShardConfig instance.
        """
        meta_dict = self.metadata.get()
        config = meta_dict.get(DLROVER_CKPT_CONFIG_KEY, default_config)
        return config


class AsyncCheckpointSaver(metaclass=ABCMeta):
    """
    CheckpointSaver asynchronously saves the state dict from the shared memory
    into the storage.

    Attributes:
        checkpoint_dir (str): the directory to save the checkpointing state
            dict to the storage if the training process fails.
        local_shard_num (int): the number of model/optimizer shards
            on the node.
        global_shard_num (int): the number of model/optimizer shards
            across all nodes.
    """

    _saver_instance = None
    _STAGE_DIR = "._dlrover_ckpt_stage"

    def __init__(
        self, checkpoint_dir, local_shard_num=1, global_shard_num=1
    ) -> None:
        self.checkpoint_dir = checkpoint_dir
        self.local_shard_num = local_shard_num
        self.global_shard_num = global_shard_num
        self._node_rank = env_utils.get_node_rank()
        self._is_agent_rank_0 = self._node_rank == 0
        self._shm_handlers: List[SharedMemoryHandler] = []
        self._shm_locks: List[SharedLock] = []

        # Indicate whether the saver is writing state to storage
        self._writing_storage = False
        # The latest step to save the checkpoint.
        self._latest_step = 0
        qname = CheckpointSharedObjPrefix.SAVE_STEP_QNAME + str(0)
        self._event_queue = SharedQueue(name=qname, create=True)
        for i in range(self.local_shard_num):
            self._shm_handlers.append(SharedMemoryHandler(i))
            lock_name = CheckpointSharedObjPrefix.SHM_LOCK_NAME + str(i)
            self._shm_locks.append(SharedLock(name=lock_name, create=True))
        self._executor = ThreadPoolExecutor(
            max_workers=self.local_shard_num, thread_name_prefix="ckpt_saver-"
        )

    def __del__(self):
        self.close()

    @classmethod
    def start_async_saving_ckpt(cls):
        """
        Start a thread to asynchronously save the checkpoint state dict
        from the shared memory into the storage. Firstly, it waits that
        the training process notify the saver class to create a saver.
        """
        sq = SharedQueue(name="factory", create=True)

        def _save():
            class_meta: SaverClassMeta = sq.get()
            module = importlib.import_module(class_meta.module_path)
            class_def = getattr(module, class_meta.class_name)
            if cls._saver_instance is None:
                saver: AsyncCheckpointSaver = class_def(**class_meta.init_args)
                cls._saver_instance = saver
            cls._saver_instance._sync_shm_to_storage()

        threading.Thread(
            target=_save, name="checkpoint-saver", daemon=True
        ).start()

    @classmethod
    def get_ckpt_saver(cls):
        return cls._saver_instance

    @classmethod
    def register_signal_handler(cls):
        sigint_handler = signal.getsignal(signal.SIGINT)
        sigterm_handler = signal.getsignal(signal.SIGTERM)

        def _clean_shm_handler(signum, frame):
            """Clean the shared memory from ^C and "killall python" etc."""
            if cls._saver_instance:
                cls._saver_instance.close()
            if callable(sigint_handler):
                sigint_handler(signum, frame)

        def _save_shm_before_exiting(signum, frame):
            """Save the state dict from the shared memory into the storage
            before the process exits.
            """
            if cls._saver_instance:
                cls._saver_instance.save_shm_to_storage()
                cls._saver_instance.close()
            if callable(sigterm_handler):
                sigterm_handler(signum, frame)

        signal.signal(signal.SIGINT, _clean_shm_handler)
        signal.signal(signal.SIGTERM, _save_shm_before_exiting)

    def close(self):
        """Clear the resource of the shared objects."""
        event = CheckpointEvent(type=CheckpointEventType.EXIT)
        if not self._event_queue.empty():
            self._event_queue._queue.get()
        self._event_queue.put(event)
        for i in range(self.local_shard_num):
            if self._shm_handlers[i]:
                self._shm_handlers[i].close()
                self._shm_handlers[i].unlink()
            self._shm_locks[i].unlink()
        self._event_queue.unlink()
        self._executor.shutdown()

    def _sync_shm_to_storage(self):
        """
        The loop to persist the state dict from the memory
        buffer into the storage.
        """
        logger.info("Async flash checkpoint saver starts!")
        while True:
            event: CheckpointEvent = self._event_queue.get()
            if (
                event.type == CheckpointEventType.UPDATE_SHARD
                and event.global_shard_num != self.global_shard_num
            ):
                logger.info(
                    "Reset the shard memory because the number of "
                    "global shards changes"
                )
                self._reset_shared_memory()
            elif event.type == CheckpointEventType.SAVE:
                logger.info(
                    f"ShardingSaver save checkpoint to storage, event {event}"
                )
                self.save_step_checkpoint(event.step)
            elif event.type == CheckpointEventType.EXIT:
                break

    def _reset_shared_memory(self):
        for shm_handler in self._shm_handlers:
            shm_handler.reset()

    def _save_shard(
        self,
        step,
        local_shard_id: int,
        ckpt_config: CheckpointShardConfig,
        step_done_dir: str,
    ):
        """Save the shard of state dict into the storage."""
        try:
            shm_handler = self._shm_handlers[local_shard_id]
            shm_lock = self._shm_locks[local_shard_id]
            if shm_handler.shared_memory is None:
                shm_handler.init_shared_memory(create=False)

            shm_lock.acquire()
            default_config = CheckpointShardConfig()
            config = shm_handler.get_checkpoint_config(default_config)
            if config.step != step:
                shm_lock.release()
                logger.error(
                    f"The step {step} in event is no equal "
                    f"to step {config.step} in memory."
                )
                return

            logger.info(
                f"Local rank {local_shard_id} Save checkpoint from the shared "
                f"memory into the storage {ckpt_config}."
            )
            self.persist_to_storage(local_shard_id, ckpt_config)
            shm_lock.release()
            global_shard_id = (
                self.local_shard_num * self._node_rank + local_shard_id
            )
            step_done_file = os.path.join(step_done_dir, str(global_shard_id))
            with open(step_done_file, "w") as f:
                f.write("done")
            return True

        except Exception as e:
            logger.error(
                f"Rank {local_shard_id} save checkpoint failed, error: {e}",
                exc_info=True,
            )
            shm_lock.release()
            return False

    def _any_rank_locked(self):
        """Verify that the shared memory of any rank is locked."""
        all_locked = []
        for lock in self._shm_locks:
            all_locked.append(lock.locked())
        return any(all_locked)

    def _get_checkpoint_done_dir(self, step):
        """Get the directory of the done files."""
        done_dir = os.path.join(
            self.checkpoint_dir, self._STAGE_DIR, str(step) + ".done"
        )
        os.makedirs(done_dir, exist_ok=True)
        return done_dir

    def save_shm_to_storage(self, timeout=120):
        """
        Save the state dict in the shared memory into the storage. The agent
        can call the method to save the state dict into the storage if the
        training process fails or the agent wants to restart training
        processes.
        """
        if any(
            [handler.no_checkpint_state() for handler in self._shm_handlers]
        ):
            logger.info(
                "Skip because no any memory buffer with the state dict."
            )
            return

        if self._writing_storage:
            logger.info("Saver is writing to storage, waiting...")
            start = time.time()
            while self._writing_storage:
                time.sleep(2)
                elapsed_time = time.time() - start
                if elapsed_time > timeout:
                    logger.error("Saver writing to storage, timeout")
                    return
        if self._any_rank_locked():
            # The training process does not release the lock because it fails
            # when writing the state dict into the shared memory. The shared
            # memory may be dirty and the saver cannot save it to the storage.
            return
        steps = []
        for shm_handler in self._shm_handlers:
            default_config = CheckpointShardConfig()
            config = shm_handler.get_checkpoint_config(default_config)
            steps.append(config.step)
        if len(set(steps)) > 1:
            logger.error(
                "Skip because steps in shards are not "
                f"inconsistent: {steps}"
            )
            return
        step = steps[0]
        if step > self._latest_step:
            self.save_step_checkpoint(step)
            logger.info(
                "Save the checkpointing state dict from the shared "
                f"memory to storage, step: {step}."
            )

    @abstractmethod
    def save_step_checkpoint(self, step: int):
        """
        Save the checkpoint of a step into the storage.

        Args:
            step (int): the iteration step.
        """
        pass

    @abstractmethod
    def persist_to_storage(self, local_shard_id, ckpt_config):
        """
        Persist the state dict to a storage path.

        Args:
            local_shard_id (int): the index of local shard.
            ckpt_config : the checkpoint config with the path to
                save the storage.
        """
        pass

    @abstractmethod
    def commit_checkpoint(self, step: int, step_done_dir: str, timeout=600):
        """
        Commit a checkpoint if all shards are saved to the storage.

        Args:
            step (int): the iteration step.
            step_done_dir (str): the directory to save the
                done file of each shard.
            timeout (int): the timeout to wait that all shards are saved,
                default 600s.
        """
        pass

    @abstractmethod
    def update_tracker_file(self, step: int):
        """
        Update the checkpoint tracker file on the storage after the
        checkpoing state dict is saved to the storage.

        Args:
            step (int): the iteration step.
        """
        pass


class CommonDirCheckpointSaver(AsyncCheckpointSaver):
    """
    The saver asynchronously save the checkpointing state dict from
    the shared memory to the storage. It launches threads to directly save
    each shard into the path configured by users. The thread saves each shard
    to a unique path and writes a done file after the saving is finshied.
    The path is generated by the training framework like Megatron-LM,
    DeepSpeed or users' definition. The saver updates the tracer file with
    the step if the number of done file is equal to the number of shard.
    """

    def update_tracker_file(self, step):
        """
        Write the step into the tracker file.

        Args:
            step (int): the checkpointing step.
        """
        tracker_filename = os.path.join(
            self.checkpoint_dir, CheckpointConstant.TRACER_FILE_NAME
        )
        with open(tracker_filename, "w") as f:
            f.write(str(step))

    def save_step_checkpoint(self, step: int):
        """
        Save the checkpoint of a step into the storage.

        Args:
            step (int): the iteration step.
        """
        self._writing_storage = True

        step_done_dir = self._get_checkpoint_done_dir(step)
        os.makedirs(step_done_dir, exist_ok=True)

        write_success = False
        # save to stage path for each local rank
        futures: List[Future] = []
        for i in range(self.local_shard_num):
            default_config = SingleFileCheckpointConfig()
            ckpt_config = self._shm_handlers[i].get_checkpoint_config(
                default_config
            )
            future: Future = self._executor.submit(
                self._save_shard,
                step,
                i,
                ckpt_config,
                step_done_dir,
            )
            futures.append(future)

        success_count = 0
        for (i, future) in enumerate(futures):
            if future.result():
                success_count += 1
            else:
                logger.error(
                    f"Fail to save checkpoint shared {i} for step {step}"
                )

        if success_count == self.local_shard_num:
            write_success = True
            self._latest_step = step

        if not write_success:
            logger.error(
                f"Rank {self._node_rank} save checkpoint failed for "
                f"step {step}"
            )
            return

        # commit checkpoint
        if self._is_agent_rank_0:
            self.commit_checkpoint(step, step_done_dir)

        self._writing_storage = False

    def commit_checkpoint(self, step: int, step_done_dir: str, timeout=600):
        """
        The node rank 0 will update the tracker file with the step
        after the number of done files is equal to the number of shard.

        Args:
            step (int): the iteration step.
            step_done_dir (str): the directory to save the
                done file of each shard.
            timeout (int): the timeout to wait that all shards are saved,
                default 600s.
        """
        start_time = time.time()
        while True:
            if len(os.listdir(step_done_dir)) == self.global_shard_num:
                # all local rank done
                self.update_tracker_file(step)

                # clean stage dir
                shutil.rmtree(step_done_dir)
                logger.info(
                    f"All agents finish saving checkpoint for step {step}"
                )
                break
            # timeout
            elapsed_time = round(time.time() - start_time, 2)
            if elapsed_time > timeout:
                logger.error(
                    f"Commit checkpoint timeout for step {step}, "
                    f"elapsed_time: {elapsed_time}"
                )
                # clean stage dir
                shutil.rmtree(step_done_dir)
                break

            time.sleep(2)


class TempDirCheckpointSaver(AsyncCheckpointSaver):
    """
    This saver firstly saves the checkpoint shards into the temporary
    directory and move the directory into the common directory configured
    by users.
    """

    def save_step_checkpoint(self, step):
        """
        Save the checkpoint of a step into the storage.

        Args:
            step (int): the iteration step.
        """
        logger.info(
            f"Rank {self._node_rank} start save checkpoint to storage, "
            f"step: {step}"
        )
        self._writing_storage = True
        ckpt_path = self.get_ckpt_path(step)

        if os.path.exists(ckpt_path):
            logger.info(f"Checkpoint for step {step} already exists, skip")
            self._writing_storage = False
            return

        stage_path = self._get_tmp_ckpt_dir(step)
        step_done_dir = self._get_checkpoint_done_dir(step)

        write_success = False
        # save to stage path for each local rank
        futures: List[Future] = []
        for i in range(self.local_shard_num):
            future = self._executor.submit(
                self._save_shard, step, i, stage_path, step_done_dir
            )
            futures.append(future)

        success_count = 0
        for (i, future) in enumerate(futures):
            if future.result():
                success_count += 1
            else:
                logger.error(
                    f"Rank {i} save checkpoint failed for step {step}"
                )

        if success_count == self.local_shard_num:
            write_success = True

        if not write_success:
            logger.error(
                f"Rank {self._node_rank} save checkpoint failed for "
                f"step {step}"
            )
            return

        # commit checkpoint
        if self._is_agent_rank_0:
            self.commit_checkpoint(
                step,
                step_done_dir=step_done_dir,
                tmp_path=stage_path,
                target_path=ckpt_path,
            )

        self._writing_storage = False

    def get_ckpt_path(self, step: int):
        """
        User can override the method to define the checkpoint name.

        Args:
            step (int): the iteration step.
        """
        name = f"{CheckpointConstant.CKPT_NAME_PREFIX}{step}"
        return os.path.join(self.checkpoint_dir, name)

    def _get_tmp_ckpt_dir(self, step: int):
        """
        Get the temp directory to save the latest checkpoint. After all
        shards are saved, the saver will move the directory to a
        regular directory defined by users.
        """
        ckpt_name = os.path.join(
            self.checkpoint_dir, self._STAGE_DIR, str(step)
        )
        os.makedirs(ckpt_name, exist_ok=True)
        return ckpt_name

    def commit_checkpoint(  # type: ignore
        self,
        step: int,
        step_done_dir: str,
        tmp_path: str,
        target_path: str,
        timeout=600,
    ):
        """
        Commit checkpoint from stage dir to target dir.

        This method is called by agent rank 0, it will check if all agent rank
        write finish, if true, it will commit checkpoint by move stage dir to
        target dir.

        Args:
            step (int): the iteration step.
            step_donr_dir (str): the directory to save the done file of
                each shard.
            tmp_path: the temp directory path to save the latest checkpoint.
            target_path: the regular diectory path to save the checkpoint.
            timeout (int): the timeout to wait all shards are saved,
                default 600s.
        """
        logger.info(
            f"Start commit checkpoint tmp_path: {tmp_path}, "
            f"path: {target_path}"
        )
        start_time = time.time()
        while True:

            # check all local rank done
            if len(os.listdir(step_done_dir)) == self.global_shard_num:
                # all local rank done
                logger.info(f"All agent done for step {tmp_path}")

                # commit checkpoint
                shutil.move(tmp_path, target_path)

                self.update_tracker_file(step)

                # clean stage dir
                shutil.rmtree(step_done_dir)
                logger.info(
                    f"Commit checkpoint tmp_path: {tmp_path}, "
                    f"path: {target_path}"
                )
                break

            # timeout
            elapsed_time = time.time() - start_time
            if elapsed_time > timeout:
                logger.error(
                    f"Commit checkpoint timeout, tmp_path: {tmp_path},"
                    f"path: {target_path}, elapsed_time: {elapsed_time}"
                )
                # clean stage dir
                shutil.rmtree(tmp_path)
                shutil.rmtree(step_done_dir)
                break

            time.sleep(2)


class DdpCheckpointSaver(CommonDirCheckpointSaver):
    """Persist the checkpoint from CPU memory buffer into the storage."""

    def persist_to_storage(
        self,
        local_shard_id: int,
        ckpt_config: SingleFileCheckpointConfig,
    ):
        if self._node_rank != 0:
            logger.info("Skip and only rank 0 saves checkpoint in a DDP job.")
            return
        state_dict = self._shm_handlers[local_shard_id].load_state_dict()
        state_dict.pop(DLROVER_CKPT_CONFIG_KEY, None)
        checkpoint_dir = os.path.dirname(ckpt_config.path)
        os.makedirs(checkpoint_dir, exist_ok=True)
        torch.save(state_dict, ckpt_config.path)


class MegatronCheckpointSaver(CommonDirCheckpointSaver):
    TRACER_FILE = "latest_checkpointed_iteration.txt"

    def update_tracker_file(self, step):
        """
        Write the step into the tracker file.

        Args:
            step (int): the checkpointing step.
        """
        tracker_filename = os.path.join(
            self.checkpoint_dir, CheckpointConstant.TRACER_FILE_NAME
        )
        with open(tracker_filename, "w") as f:
            f.write(str(step))
        ds_tracker_filename = os.path.join(
            self.checkpoint_dir, self.TRACER_FILE
        )
        with open(ds_tracker_filename, "w") as f:
            f.write(str(step))

    def persist_to_storage(
        self,
        local_shard_id: int,
        ckpt_config: SingleFileCheckpointConfig,
    ):
        state_dict = self._shm_handlers[local_shard_id].load_state_dict()
        state_dict.pop(DLROVER_CKPT_CONFIG_KEY, None)
        checkpoint_dir = os.path.dirname(ckpt_config.path)
        os.makedirs(checkpoint_dir, exist_ok=True)
        torch.save(state_dict, ckpt_config.path)


class DeepSpeedCheckpointSaver(CommonDirCheckpointSaver):
    TRACER_FILE = "latest"

    def update_tracker_file(self, step):
        """
        Write the step into the tracker file.

        Args:
            step (int): the checkpointing step.
        """
        tracker_filename = os.path.join(
            self.checkpoint_dir, CheckpointConstant.TRACER_FILE_NAME
        )
        with open(tracker_filename, "w") as f:
            f.write(str(step))
        ds_tracker_filename = os.path.join(
            self.checkpoint_dir, self.TRACER_FILE
        )
        with open(ds_tracker_filename, "w") as f:
            f.write(str(step))

    def persist_to_storage(  # type: ignore
        self,
        local_shard_id: int,
        ckpt_config: DeepSpeedCheckpointConfig,
    ):
        """Persist the checkpoint from CPU memory buffer into the storage."""
        state_dict = self._shm_handlers[local_shard_id].load_state_dict()
        state_dict.pop(DLROVER_CKPT_CONFIG_KEY, None)
        model_sd = state_dict.get(CheckpointConstant.MODEL_STATES_NAME, {})
        if model_sd and ckpt_config.model_path:
            checkpoint_dir = os.path.dirname(ckpt_config.model_path)
            os.makedirs(checkpoint_dir, exist_ok=True)
            torch.save(model_sd, ckpt_config.model_path)

        optimizer_sd = state_dict.get(CheckpointConstant.OPTIM_STATES_NAME, {})
        if optimizer_sd and ckpt_config.optimizer_path:
            checkpoint_dir = os.path.dirname(ckpt_config.optimizer_path)
            os.makedirs(checkpoint_dir, exist_ok=True)
            torch.save(optimizer_sd, ckpt_config.optimizer_path)


class FsdpDcpSaver(CommonDirCheckpointSaver):
    """The saver saves the distributed checkpoint of FSDP into the storage."""

    def persist_to_storage(
        self,
        local_shard_id: int,
        ckpt_config: SingleFileCheckpointConfig,
    ):
        """
        Persist the state dict to a storage path.

        Args:
            local_shard_id (int): the index of local shard.
            ckpt_config : the checkpoint config with the path to
                save the storage.
        """
        shm_handler = self._shm_handlers[local_shard_id]
        checkpoint_dir = os.path.dirname(ckpt_config.path)
        os.makedirs(checkpoint_dir, exist_ok=True)
        with open(ckpt_config.path, "wb") as stream:
            assert shm_handler.shared_memory is not None
            stream.write(shm_handler.shared_memory.buf)
            os.fsync(stream.fileno())
        if self._is_agent_rank_0 and local_shard_id == 0:
            parent_path = Path(os.path.dirname(ckpt_config.path))
            meta_dict = shm_handler.metadata.get()
            dcp_metadata = meta_dict.get("dcp_metadata", {})
            if dcp_metadata:
                with (parent_path / ".metadata.tmp").open("wb") as stream:
                    pickle.dump(dcp_metadata, stream)
                    os.fsync(stream.fileno())
                (parent_path / ".metadata.tmp").rename(
                    parent_path / ".metadata"
                )
            tracer_file = os.path.join(
                self.checkpoint_dir, CheckpointConstant.TRACER_FILE_NAME
            )
            with open(tracer_file, "w") as f:
                f.write(str(ckpt_config.step))
